/**
 * Author: Andreas Gasper
 */

:- multifile query_description/3.
:- multifile metric_/3.
:- multifile metric_attributes/2.

query_description(cyclomatic_complexity, metric, 'Number of linear indipendent path through the control flow').
metric_attributes(cyclomatic_complexity, int(range(0,infinity),defaultthresholds(0,10))).

/**
 *	metric_(cyclomatic_complexity, -MethodID, -Complexity)
 *	
 *	Gives the MethodID and the cyclomatic Complexity of methods.
 *	The cyclomatic complexity is defines as: 
 *	cyclomatic_complexity = edges - nodes + 2
 *	of a controlflow graph. 
 */

metric_(cyclomatic_complexity, MethodID, Complexity):-
	internal_method(MethodID),
    number_of_statements_in_a_method(MethodID, Lines),
    \+(Lines == 0),
    getAllEdges(MethodID, Edges),
    checkLast(MethodID, Addon),
    Complexity is Edges - Lines + 2 + Addon
    .

metric_(cyclomatic_complexity, MethodID, 1):-
	internal_method(MethodID),
    number_of_statements_in_a_method(MethodID, 0).


/**
 *	getAllEdges(?MethodID, -Edges)
 *	
 *	Edges is the number of edges in the controlflowgraph of a Method
 */

getAllEdges(MethodID, Edges):-
	%The edges from the block-statements
	blockEdges(MethodID, BlockEdges),
    %The additional edges of the loop-statements
    loopEdges(MethodID, LoopEdges),
    %The additional edges of the if-statements   
    ifEdges(MethodID, IfEdges),
    %The additional edges of the switch-statements   
    switchEdges(MethodID, SwitchEdges),
    %The additional edges of the catch-statements
    tryCatchEdges(MethodID, TryCatchEdges),
    %Adding everything up  
    Edges is BlockEdges + LoopEdges + IfEdges + SwitchEdges + TryCatchEdges
    .
    
/**
 *	blockEdges(+MethodID, -AllBlockEdges)
 *	
 *	AllBlockEdges is the sum of edges in all blockTs of a Method.
 */  
 
blockEdges(MethodID, AllBlockEdges):-    
    findall(LengthOfBlock,
    	(
    		ri_blockT_encl(MethodID, BlockID),
    		blockT(BlockID, _, _, Statements),
    		length(Statements, LengthOfBlock)
    	),
    	AllBlocks),
    %The number of edges for a block is the length of the statementslist - 1.
    %For all blocks together its the sum of all statements of all blocks minus the number of blocks 
    length(AllBlocks, Discount),
    sumlist(AllBlocks, BlockEdges),
    AllBlockEdges is BlockEdges - Discount
    .

/**
 * loopEdges(+MethodID, -LoopEdges)
 *
 * LoopEdges is the number of the loop statements of a method times 2.
 */

loopEdges(MethodID, LoopEdges):-     
    findall(Loop,
    	(
    		ri_doWhileT_encl(MethodID, Loop);
    		ri_forT_encl(MethodID, Loop);
    		ri_foreachT_encl(MethodID, Loop);
    		ri_whileT_encl(MethodID, Loop)
    	),
    	AllLoops),
    length(AllLoops, NumberOfLoops),
    LoopEdges is NumberOfLoops * 2
    . 

/**
 * ifEdges(+MethodID, -IfEdges)
 *
 * IfEdges is the number of all if statements * 2 + the number of all ELSE statements of the Method
 * 											 IF	O						 IF	O		
 *                 								| \						  /   \
 * 												| Then					Then Else
 * 												| /						  \   /
 * 												O							O  
 */

ifEdges(MethodID, IfEdges):-    
    findall(If,
    	(
    		ri_ifT_encl(MethodID, If)
    	),
    	Ifs),
    length(Ifs, NumberOfIfs),
    IfThenAddition is NumberOfIfs * 2,
    %This findall finds all ELSE parts and then adds then number of else parts to the IfThenAddition 
    findall(Else,
    	(
    		ri_ifT_encl(MethodID, IfID),
    		ifT(IfID, _, _, _, _, Else),
    		\+(Else = null) 
    	),
    	Elses),
    length(Elses, NumberOfElses),
    IfEdges is IfThenAddition + NumberOfElses
    .

/**
 * switchEdges(+MethodID, -SwitchEdges)
 *
 * SwitchEdges is the number of edges in all switch blocks of the Method.
 */
   
switchEdges(MethodID, SwitchEdges):-
	findall(SwitchEdgeNumber,	
		(
			ri_switchT_encl(MethodID, SwitchID),
			switchT(SwitchID, _, MethodID, _, Statements),
			findall(CaseID,	
				(
					 ri_caseT_parent(SwitchID, CaseID)
				),
				CaseIDs),
			length(Statements, NumberOfStatements),
			length(CaseIDs, NumberOfCases),
			SwitchEdgeNumber is NumberOfCases + NumberOfStatements
		),
		SwitchEdgeNumbers),
	sumlist(SwitchEdgeNumbers, SwitchEdgess),
	length(SwitchEdgeNumbers, Discount),
	SwitchEdges is SwitchEdgess - Discount
	.

/**
 * catchEdges(+MethodID, -CatchEdges)
 *
 * CatchEdges is the number of the catch statements of a method times 2.
 */
tryCatchEdges(MethodID, TryCatchEdges):-
    findall(CatchEdgeNumber,
    	(
    		ri_catchT_encl(MethodID, CatchEdgeNumber)
    	),
    	CatchEdgesNumber),
    length(CatchEdgesNumber, NumberOfCatchEdges),
    CatchEdges is NumberOfCatchEdges * 2,
    findall(TryEdgeNumber,
    	(
    		ri_tryT_encl(MethodID, TryEdgeNumber)
    	),
    	TryEdgesNumber),
    length(TryEdgesNumber, NumberOfTryEdges),
    TryEdges is NumberOfTryEdges * 2,
    TryCatchEdges is CatchEdges + TryEdges
    .

/**
 * checkLast(+MethodID, -Addon)
 *
 * If the last statement of the method is an if or switch statement we have to add
 * a additonal end node in the controlflow graph. Therefor if the last statement is
 * an if or switch statement Addon is unified with 1.
 */
checkLast(MethodID, Addon):-
	ri_blockT_encl(MethodID, BlockID),
	blockT(BlockID, _, _, Statements),
	last(Statements, ID),
	(
		  ifT(ID, MethodID, _, _, _, _);
		  switchT(ID, MethodID, _, _, _);
		  catchT(ID, MethodID, _, _, _)
	),
	! ,
	Addon is 1 
	.

checkLast(_MethodID, 0).
