:- dynamic dependency_db/4.             % dependency_db(p1,  p3, packageDep, [weight]).
:- dynamic package_db/2.                % package_db(packageId,packageName).
:- dynamic package_dependency_db/4.

/**
       
 packageDependency(-DependentId, -DependeeId,-SccNumber,-DependencyDegree)
        Packages = [[pack1Id,pack1Name],...]
        Dependencies = [[ReferencingPackageId,DependendPackageID,IsPartOfCircleNo,DegreeOfDependency],...]
        
        If there are no dependencies, there will be an empty list.
        If there are no packages, this would'nt trigger
        
        IsPartOfCircleNo   : All strongly connected components are assigned a number. This is the
                             number for the scc, this edge is part of.
        DegreeOfDependency : Number of classes in ReferencingPackageId that depend on DependenPackageID.
        
*/
packageDependency(DependentId, DependeeId, SccNumber, DependencyDegree):-
      packageDependency(DepList),
      member([DependentId,DependeeId,SccNumber,DependencyDegree],DepList).

packageDependency(DepList) :-
        build_base,
%        packageslist_with_names(Packages),
        dependencieslist(DepList),
        delete_base.

      


/**     
        build_base
        
        removes all dependencies and packages from the factbase
        adds the dependency_db/4 and package_db/2 predicates to the factbase    
*/
build_base :-
        delete_base,
        add_packages,
        add_dependencies.

/**     
        delete_base
        
        removes the dependency_db/4 ,package_dependency_db/4 and package_db/2 predicates from the factbase      
*/
delete_base :-
        retractall(package_db(_,_)),
        retractall(dependency_db(_,_,_,_)),
        retractall(package_dependency_db(_,_,_,_)).

/**     
        add_packages
        
        adds package_db/2 predicates to the factbase.
*/

add_packages :-
    packageT(Id,Name),
    not(external_package(Id)),
    assert(package_db(Id,Name)),
    fail.
add_packages.

/**     
        add_dependencies
        
        adds dependency_db/4 predicates to the factbase.
*/

add_dependencies :-
    package_db(DependingPackage,_),
    package_db(ReferencedPackage,_),

    not(DependingPackage = ReferencedPackage),
    checkDependencyBetweenPackages(DependingPackage,ReferencedPackage,Weight),
    % Here we use/define the convention that package dependencies are marked 
    % by 'packageDep' in the third argument of dependency_db/4:
    
    assert(dependency_db(DependingPackage,ReferencedPackage,packageDep, [Weight])),
    fail.
add_dependencies.


    
/**
        packageslist(-Packages)
        
        Will bind "Packages" to a list of all not extern packages available
        in the factbase
*/
packageslist(Packages) :-
    findall(PackageId,package_db(PackageId,_),Packages).

packageslist_with_names(Packages) :-
    findall([PackageId,PackageName],package_db(PackageId,PackageName),Packages).

/**
 * dependencieslist(Dependencies)
 */     
dependencieslist(Dependencies) :-
    packageslist(PackageIds),
    max_cycles(PackageIds,_,StronglyConnectedComponents),
    %write(StronglyConnectedComponents),
    store_all_cycle_edges(StronglyConnectedComponents),
    store_acyclic_edges,
    findall([ReferencingPackageId,DependendPackageId,CircleNo,Weight],
                package_dependency_db(ReferencingPackageId,DependendPackageId,CircleNo, Weight),
                Dependencies),
    !.                  
%Just for the case that max_cycles fails, if there are no components...
dependencieslist(Dependencies) :-
    store_acyclic_edges,
    findall([ReferencingPackageId,DependendPackageId,CircleNo,Weight],
                package_dependency_db(ReferencingPackageId,DependendPackageId,CircleNo, Weight),
                Dependencies),
    !.
%
dependencieslist([[]]).



/**
        store_all_cycle_edges(+StronglyConnectedComponents)
        
        e.g. StronglyConnectedComponents = [[1,2,3],[4,5,6]]
        This will store all edges of all StronglyConnectedComponents 
        given to the factbase using the package_dependency_db predicate
*/
store_all_cycle_edges(L) :-
    store_all_cycle_edges(L,1).

store_all_cycle_edges([],_).

store_all_cycle_edges([FirstCycle|Rest],Depth) :-
   store_cycle_edges(FirstCycle,Depth),
   Deeper is Depth+1,
   store_all_cycle_edges(Rest,Deeper).
   
store_cycle_edges(Cycle,Depth) :-
    store_cycle_edges(Cycle,Cycle,Depth) .

store_cycle_edges([],_Cycle,_Depth) .

store_cycle_edges([Node|Rest],Cycle,CycleNo) :-
    remember_dependency(Node,Cycle,CycleNo),
    store_cycle_edges(Rest,Cycle,CycleNo).


/**
        remember_dependency(+Node,+Cycle,+CycleNo)
        
        This will store all dependencies from the node "Node" to any other 
        node from the list "Cycle" to the package_dependency_db predicate.
        Also given is "CycleNo", the number of the actual strongly connected
        component
*/
remember_dependency(Node,Cycle,CycleNo) :-
    dependency_db(Node,Other,packageDep, [Weight]),
      member(Other,Cycle),
      assert( package_dependency_db(Node,Other,CycleNo,Weight) ),
    fail.
remember_dependency(_Node,_Cycle,_CycleNo).


/**
        store_acyclic_edges
        
        This will store all edges to the package_dependency_db predicate which
        are nocht member of a cycle. The cycle number will be 0.
*/

store_acyclic_edges :-
    dependency_db(ReferencingPackageId,DependendPackageId,packageDep, [Weight]),
      not(package_dependency_db(ReferencingPackageId,DependendPackageId,_,_)),
      assert( package_dependency_db(ReferencingPackageId,DependendPackageId,0, Weight) ),
    fail. 
store_acyclic_edges.  
        



/**
        checkDependencyBetweenPackages(+DependingPackage,+ReferencedPackage,?ResultWeight)
        
        cecks if there is a dependency from 'DependingPackage' 
        to 'ReferencedPackage'. 
        ResultWeight will be bound to the number of classes connecting from 'DependingPackage'
        to 'ReferencedPackage'.
*/ 
 
checkDependencyBetweenPackages(ReferencedPackage,ReferencedPackage,0) :- 
    nonvar(ReferencedPackage),
    !.

checkDependencyBetweenPackages(DependingPackage,ReferencedPackage,ResultWeight) :-
    nonvar(ReferencedPackage),
    nonvar(DependingPackage),

        findall(dependsOn(DependingPackage,ReferencedPackage),
                depends(DependingPackage,ReferencedPackage),
                List),
        length(List,ResultWeight),
        not(ResultWeight = 0).
    

/**
        depends(?DependingPackage,?ReferencedPackage)
    
    Will succeed if there is any kind of connection/depending
    between DependingPackage and ReferencedPackage
*/
depends(DependingPackage,ReferencedPackage) :-
    ri_compilationUnitT_parent(DependingPackage, DependingCompilationUnit),
    ri_classT_parent(DependingCompilationUnit, DependingClassID),
    ri_compilationUnitT_parent(ReferencedPackage, ReferencedCompilationUnit),
    ri_classT_parent(ReferencedCompilationUnit, ReferencedClassID),
    not( DependingPackage = ReferencedPackage ),   %%% <-- Don't forget this! - GK
    % not(externT(ReferencedClassID)),
   	classDependency(DependingClassID, ReferencedClassID, _).
   	